@@ -3,7 +3,6 @@
 #  Use of this source code is governed by a BSD-3-clause license that can be
 #  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause
 
-import six
 import logging
 from coremltools.converters.mil.input_types import (
     InputType,
@@ -157,15 +156,12 @@ class TFConverter:
                 if inputs[0].name is None:
                     inputs[0].name = tf_placeholder_names[0]
 
-            # filter out those inputs which is not in tf_placeholder_names
-            inputs = [x for x in inputs if x.name in tf_placeholder_names]
-
             # We fill in shapes for user-specified input that doesn't have shape
             for inp in inputs:
                 # Check inputs existence
                 if inp.name is None:
                     raise ValueError(
-                        "Unable to infer input's name or input name was not provided"
+                        "Multiple inputs are found in graph, but no input name was provided"
                     )
                 if inp.name not in tf_placeholder_names:
                     raise ValueError(
@@ -228,7 +224,7 @@ class TFConverter:
         self._validate_outputs(tfssa, outputs)
         outputs = main_func.outputs if outputs is None else outputs
         outputs = outputs if isinstance(outputs, (tuple, list)) else [outputs]
-        outputs = [x if isinstance(x, six.string_types) else x.name for x in outputs]
+        outputs = [x if isinstance(x, str) else x.name for x in outputs]
         self.outputs = outputs
 
         # We would like a stack so that we run conversion sequentially.
@@ -279,7 +275,7 @@ class TFConverter:
     @staticmethod
     def _get_tensor_name(tensor):
         ret = None
-        if isinstance(tensor, six.string_types):
+        if isinstance(tensor, str):
             ret = tensor
         else:
             ret = tensor.name
