@@ -7,8 +7,7 @@ from coremltools.converters.mil import testing_reqs
 from coremltools.converters.mil.testing_reqs import *
 from coremltools.converters.mil.frontend.tensorflow.test.testing_utils import (
     make_tf_graph,
-    tf_graph_to_mlmodel,
-    run_compare_tf,
+    tf_graph_to_mlmodel
 )
 
 # Custom Op imports
@@ -36,10 +35,10 @@ class TestCustomMatMul:
         input_spec = InputSpec(
             x=TensorInputType(),
             y=TensorInputType(),
-            transpose_x=BoolInputType(const=True, default=False),
-            transpose_y=BoolInputType(const=True, default=False),
-            x_is_sparse=BoolInputType(const=True, default=False),
-            y_is_sparse=BoolInputType(const=True, default=False),
+            transpose_x=BoolInputType(const=True, optional=True),
+            transpose_y=BoolInputType(const=True, optional=True),
+            x_is_sparse=BoolInputType(const=True, optional=True),
+            y_is_sparse=BoolInputType(const=True, optional=True),
         )
 
         # Specifying binding for custom op for specifying inputs,
@@ -51,6 +50,14 @@ class TestCustomMatMul:
             "description": "Custom Sparse MatMul Layer",
         }
 
+        def default_inputs(self):
+            return DefaultInputs(
+                transpose_x=False,
+                transpose_y=False,
+                x_is_sparse=False,
+                y_is_sparse=False,
+                )
+
         def __init__(self, **kwargs):
             super(TestCustomMatMul.custom_sparse_matmul, self).__init__(**kwargs)
 
@@ -87,19 +94,28 @@ class TestCustomMatMul:
 
     @pytest.mark.skipif(not testing_reqs._HAS_TF_1, reason=MSG_TF1_NOT_FOUND)
     @pytest.mark.parametrize(
-        "use_cpu_only, backend, transpose_a, transpose_b," "a_is_sparse, b_is_sparse",
+        "use_cpu_only, backend, transpose_a, transpose_b," "a_is_sparse, b_is_sparse, b_is_const",
         itertools.product(
-            [True], backends, [True, False], [True, False], [True, False], [True, False]
+            [True], backends, [True, False], [True, False], [True, False], [True, False], [True, False],
         ),
     )
     def test_tf(
-        self, use_cpu_only, backend, transpose_a, transpose_b, a_is_sparse, b_is_sparse
+        self, use_cpu_only, backend, transpose_a, transpose_b, a_is_sparse, b_is_sparse, b_is_const,
     ):
+        if backend[0] == 'mlprogram':
+            pytest.xfail("Custom layer not supported with ML Program backend")
+
         rank = 2
         shape = list(np.random.randint(low=3, high=100, size=1)) * rank
         with tf.Graph().as_default() as graph:
             x = tf.placeholder(tf.float32, shape=shape)
-            y = tf.placeholder(tf.float32, shape=shape)
+
+            if b_is_const:
+                y_value = random_gen(shape, rand_min=-100, rand_max=100)
+                y = tf.constant(y_value, dtype=tf.float32)
+            else:
+                y = tf.placeholder(tf.float32, shape=shape)
+
             ref = tf.sparse_matmul(
                 x,
                 y,
@@ -112,6 +128,8 @@ class TestCustomMatMul:
                 graph,
                 {
                     x: random_gen(shape, rand_min=-100, rand_max=100),
+                } if b_is_const else {
+                    x: random_gen(shape, rand_min=-100, rand_max=100),
                     y: random_gen(shape, rand_min=-100, rand_max=100),
                 },
                 ref,
@@ -135,6 +153,8 @@ class TestCustomMatMul:
                 b_is_sparse == layers[-1].custom.parameters["y_is_sparse"].boolValue
             ), "Incorrect parameter value k"
 
+            assert len(layers) == 2 if b_is_const else len(layers) == 1
+
 
 class TestCustomTopK:
     @pytest.fixture(scope="class")
@@ -144,9 +164,9 @@ class TestCustomTopK:
         class custom_topk(Operation):
             input_spec = InputSpec(
                 x=TensorInputType(),
-                k=IntInputType(const=True, default=1),
-                axis=IntInputType(const=True, default=-1),
-                sorted=BoolInputType(const=True, default=False),
+                k=IntInputType(const=True, optional=True),
+                axis=IntInputType(const=True, optional=True),
+                sorted=BoolInputType(const=True, optional=True),
             )
 
             bindings = {
@@ -156,6 +176,13 @@ class TestCustomTopK:
                 "description": "Top K Custom layer",
             }
 
+            def default_inputs(self):
+                return DefaultInputs(
+                    k=1,
+                    axis=-1,
+                    sorted=False,
+                    )
+
             def __init__(self, **kwargs):
                 super(custom_topk, self).__init__(**kwargs)
 
@@ -173,7 +200,6 @@ class TestCustomTopK:
                 ret_shape[axis] = k
                 return types.tensor(x_type, ret_shape), types.tensor(types.int32, ret_shape)
 
-        # TODO: rdar://61241807 ([MIL] [Polish] Custom layer operator documentation)
         # Following logging is to ensure testing of TopK implemented in tf converter
         # default path is testing with appropriate conversion function
         # Log default tf topk
@@ -200,6 +226,9 @@ class TestCustomTopK:
     )
     @pytest.mark.usefixtures("create_custom_TopK")
     def test_tf(self, use_cpu_only, backend, rank, k):
+        if backend[0] == 'mlprogram':
+            pytest.xfail("Custom layer not supported with ML Program backend")
+
         shape = np.random.randint(low=3, high=6, size=rank)
         with tf.Graph().as_default() as graph:
             x = tf.placeholder(tf.float32, shape=shape)
