@@ -2,21 +2,22 @@
 #
 #  Use of this source code is governed by a BSD-3-clause license that can be
 #  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause
-
-import six
-from coremltools import TensorType
-import pytest
 import numpy as np
+import os
+import pytest
+import tempfile
+
+from coremltools import TensorType
+import coremltools.models.utils as coremltoolsutils
+from coremltools.converters.mil.testing_utils import compare_shapes, \
+    compare_backend, run_core_ml_predict, ct_convert
+from coremltools.converters.mil.testing_reqs import ct
 
 tf = pytest.importorskip("tensorflow", minversion="1.14.0")
-from coremltools.converters.mil.testing_utils import compare_shapes, compare_backend
-from coremltools.converters.mil.testing_reqs import ct
+
 from tensorflow.python.framework import dtypes
-import tempfile
-import os
 from tensorflow.python.tools.freeze_graph import freeze_graph as freeze_g
 from tensorflow.python.keras.saving import saving_utils as _saving_utils
-import numpy as np
 
 frontend = "tensorflow"
 
@@ -76,8 +77,8 @@ def get_tf_keras_io_names(model):
     except:
         for o in model.outputs:
             output_names.append(o.name.split(":")[0].split("/")[-1])
-    for i in model.inputs:
-        input_names.append(i.name.split(":")[0])
+    for name in model.input_names:
+        input_names.append(name.split(":")[0])
     return input_names, output_names
 
 
@@ -96,7 +97,7 @@ def get_tf_node_names(tf_nodes, mode="inputs"):
         tf_nodes = [tf_nodes]
     names = list()
     for n in tf_nodes:
-        tensor_name = n if isinstance(n, six.string_types) else n.name
+        tensor_name = n if isinstance(n, str) else n.name
         if mode == "outputs":
             names.append(tensor_name)
             continue
@@ -111,7 +112,8 @@ def get_tf_node_names(tf_nodes, mode="inputs"):
 
 
 def tf_graph_to_mlmodel(
-    graph, feed_dict, output_nodes, frontend="tensorflow", backend="nn_proto"
+    graph, feed_dict, output_nodes, frontend="tensorflow", backend=("neuralnetwork", "fp32"),
+    use_cpu_for_conversion=False,
 ):
     """
     Parameters
@@ -126,6 +128,10 @@ def tf_graph_to_mlmodel(
         Frontend to convert from.
     backend: str
         Backend to convert to.
+    use_cpu_for_conversion: bool
+        Argument which is passed as is to the unified converter API.
+        That is, "ct.convert(...., useCPUOnly=use_cpu_for_conversion)"
+        It forces the model to be loaded on the CPU context, post conversion.
     -----------
     Returns MLModel, Input Values, Output Names
     """
@@ -139,8 +145,9 @@ def tf_graph_to_mlmodel(
     output_names = get_tf_node_names(output_nodes, mode="outputs")
     input_values = {name: val for name, val in zip(input_names, feed_dict.values())}
 
-    mlmodel = ct.convert(
-        graph, inputs=None, outputs=output_names, source=frontend, convert_to=backend
+    mlmodel = ct_convert(
+        graph, inputs=None, outputs=output_names, source=frontend, convert_to=backend,
+        useCPUOnly=use_cpu_for_conversion,
     )
 
     return mlmodel, input_values, output_names, output_nodes
@@ -169,9 +176,10 @@ def run_compare_tf(
     feed_dict,
     output_nodes,
     use_cpu_only=False,
+    use_cpu_for_conversion=False,
     frontend_only=False,
     frontend="tensorflow",
-    backend="nn_proto",
+    backend=("neuralnetwork", "fp32"),
     atol=1e-04,
     rtol=1e-05,
     validate_shapes_only=False,
@@ -191,6 +199,19 @@ def run_compare_tf(
         List of names representing outputs.
     use_cpu_only: bool
         If true, use CPU only for prediction, otherwise, use GPU also.
+    use_cpu_for_conversion: bool
+        If true, the converter is invoked using "ct.convert(...., useCPUOnly=True)",
+        which in turn forces the model to be loaded with the CPU context, which happens
+        when the converter loads the ML model object from the proto spec
+        using "ct.models.MLModel(proto_spec, useCPUOnly=True)".
+        The other argument, i.e., "use_cpu_only" on the other hand refers to only the compute engine
+        for prediction purposes. For a model that is loaded on a non-CPU context, it can still be forced
+        to execute on the CPU at the time of prediction. Hence,
+        "use_cpu_for_conversion = False && use_cpu_only = True" is valid and results in a case when a model is
+        loaded for GPU but executed on the CPU.
+        The scenario, "use_cpu_for_conversion = True && use_cpu_only = False" is invalid though,
+        since once a model is loaded on a CPU context its context cannot be changed to a non CPU device
+        at the time of prediction.
     frontend_only: bool
         If true, skip the prediction call, only validate conversion.
     frontend: str
@@ -207,14 +228,18 @@ def run_compare_tf(
         If present, use it as TensorFlow predictions
 
     Return:
-        Proto
+        Proto, mlmodel, input dictionay, prediction(if possible)
     """
+    if use_cpu_for_conversion and not use_cpu_only:
+        # use_cpu_for_conversion = True && use_cpu_only = False
+        raise ValueError("use_cpu_for_conversion = True && use_cpu_only = False is an invalid test case")
+
     mlmodel, input_key_values, output_names, output_nodes = tf_graph_to_mlmodel(
-        graph, feed_dict, output_nodes, frontend, backend
+        graph, feed_dict, output_nodes, frontend, backend, use_cpu_for_conversion=use_cpu_for_conversion,
     )
 
-    if frontend_only:
-        return
+    if frontend_only or coremltoolsutils._macos_version() < (10, 13):
+        return mlmodel._spec, mlmodel, input_key_values, None
 
     if not isinstance(output_nodes, (tuple, list)):
         output_nodes = [output_nodes]
@@ -273,9 +298,14 @@ def run_compare_tf(
             atol=atol,
             rtol=rtol,
             also_compare_shapes=True,
+            dtype=backend[1],
         )
-
-    return mlmodel._spec
+    pred=None
+    if not coremltoolsutils._has_custom_layer(mlmodel.get_spec()):
+        pred = run_core_ml_predict(mlmodel, input_key_values, use_cpu_only)
+    else:
+        print('Skipping model prediction as it has a custom nn layer!')
+    return mlmodel._spec, mlmodel, input_key_values, pred
 
 
 def layer_counts(spec, layer_type):
@@ -293,3 +323,37 @@ def layer_counts(spec, layer_type):
         if layer.WhichOneof("layer") == layer_type:
             n += 1
     return n
+
+
+class TensorFlowBaseTest(object):
+    testclassname=''
+    testmodelname=''
+    @pytest.fixture(autouse=True)
+    def store_testname_with_args(self, request):
+        TensorFlowBaseTest.testclassname = type(self).__name__
+        TensorFlowBaseTest.testmodelname = request.node.name
+
+    def teardown_method(self, method):
+        pass
+
+    @staticmethod
+    def run_compare_tf(graph, feed_dict, output_nodes, use_cpu_only=False,
+                       use_cpu_for_conversion=False,
+                       frontend_only=False, frontend="tensorflow",
+                       backend=("neuralnetwork", "fp32"), atol=1e-04, rtol=1e-05,
+                       validate_shapes_only=False, freeze_graph=False,
+                       tf_outputs=None):
+        res = run_compare_tf(graph, feed_dict, output_nodes,
+                        use_cpu_only=use_cpu_only,
+                        use_cpu_for_conversion=use_cpu_for_conversion,
+                       frontend_only=frontend_only, frontend=frontend,
+                       backend=backend, atol=atol,
+                       rtol=rtol,
+                       validate_shapes_only=validate_shapes_only,
+                       freeze_graph=freeze_graph, tf_outputs=tf_outputs)
+        alist = []
+        if res is not None:
+            alist = list(res)
+        alist.append(TensorFlowBaseTest.testclassname)
+        alist.append(TensorFlowBaseTest.testmodelname)
+        return tuple(alist)
