@@ -3,9 +3,7 @@
 #  Use of this source code is governed by a BSD-3-clause license that can be
 #  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause
 
-from __future__ import print_function as _
 
-from six import string_types as _string_types
 import logging as _logging
 import torch as _torch
 from coremltools._deps import version_lt
@@ -25,8 +23,11 @@ from .internal_graph import *
 from .ops import *
 from .torch_op_registry import _TORCH_OPS_REGISTRY
 from .torchir_passes import *
+from .ssa_passes.torch_passes import torch_passes
 
 torch_to_mil_types = {
+    _torch.bool: types.bool,
+    _torch.float16: types.fp16,
     _torch.float32: types.fp32,
     _torch.float64: types.fp64,
     _torch.int32: types.int32,
@@ -49,9 +50,9 @@ class TranscriptionContext:
     def add(self, ssa_var, torch_name=None):
         """
         Arguments:
-            ssa_var: Varable to add to the graph being constructed.
+            ssa_var: Variable to add to the graph being constructed.
             torch_name: Optional unique string identifier of the operation. If
-                ommitted, it will use @ssa_var.name.
+                omitted, it will use @ssa_var.name.
         """
         if torch_name is None:
             torch_name = ssa_var.name
@@ -73,6 +74,10 @@ class TranscriptionContext:
             "Torch var {} not found in context {}".format(torch_name, self.name)
         )
 
+    def __contains__(self, torch_name):
+        """Returns whether or not the torch var exist in context."""
+        return torch_name in self._current_graph[-1]
+
     def push(self, inputs=None):
         """
         Add another frame to the context. Optionally provide a tuple of
@@ -147,10 +152,12 @@ class TorchConverter:
             transform_inplace_ops,
             flatten_graph_input_values,
             flatten_graph_output_values,
+            generate_tensor_assignment_ops,
         ]
         for p in passes:
             p(self.graph)
         self.inputs = [v for v in self.graph.inputs.values()]
+        self.torch_passes = torch_passes
 
     @staticmethod
     def _check_ops(graph):
@@ -216,14 +223,22 @@ class TorchConverter:
             ):
                 self.context.add(ssa_func.inputs[users_name], torch_name=internal_name)
             for name, val in self.graph.params.items():
-                mode = decide_immediate_or_file(val)
-                const = mb.const(val=val, mode=mode, name=name)
+                const = mb.const(val=val, name=name)
                 self.context.add(const)
 
             # Add the rest of the operations
             convert_nodes(self.context, self.graph)
 
             graph_outputs = [self.context[name] for name in self.graph.outputs]
+
+            # An output can be None when it's a None constant, which happens
+            # in Fairseq MT.
+            for g in graph_outputs:
+                if g is None:
+                    msg = "Droping output {} which is None"
+                    _logging.warning(msg.format(g))
+            graph_outputs = [g for g in graph_outputs if g is not None]
+
             # Output renaming occurs
             if self.output_names:
                 for index, var in enumerate(graph_outputs):
@@ -232,9 +247,7 @@ class TorchConverter:
 
             ssa_func.set_outputs(graph_outputs)
             prog.add_function("main", ssa_func)
-
-        # TODO (sberardi): graph cleanup passes
-        # rdar://60177439
+        self.torch_passes(prog)
         return prog
 
     @staticmethod
