@@ -2,16 +2,25 @@
 #
 #  Use of this source code is governed by a BSD-3-clause license that can be
 #  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause
+import math
+import numpy as np
 
-import scipy
-from ._op_reqs import *
-
-"""
-Elementwise Unary Op Superclass
-"""
-
+from coremltools.converters.mil.mil import Operation, types, VALUE, SYMBOL
+from coremltools.converters.mil.mil.types.symbolic import is_symbolic
+from coremltools.converters.mil.mil.operation import precondition
+from coremltools.converters.mil.mil.ops.defs._op_reqs import register_op
+from coremltools.converters.mil.mil.input_type import (
+    DefaultInputs,
+    FloatInputType,
+    InputSpec,
+    ScalarOrTensorInputType,
+    StringInputType
+)
 
 class elementwise_unary(Operation):
+    """
+    Elementwise Unary Op Superclass
+    """
     input_spec = InputSpec(x=ScalarOrTensorInputType(),)
 
     def __init__(self, **kwargs):
@@ -20,25 +29,23 @@ class elementwise_unary(Operation):
     def type_inference(self):
         return self.x.sym_type
 
-
 """
 Elementwise unary op implementation(s)
 """
 
-
 @register_op(doc_str="")
 class abs(elementwise_unary):
     """
-    Returns the absolute values of the input ``x``, element-wise.
+    Return the absolute values of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -56,16 +63,16 @@ class abs(elementwise_unary):
 @register_op(doc_str="")
 class acos(elementwise_unary):
     """
-    Returns the inverse cosine values of the input ``x``, element-wise.
+    Return the inverse cosine values of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -83,16 +90,16 @@ class acos(elementwise_unary):
 @register_op(doc_str="")
 class asin(elementwise_unary):
     """
-    Returns the inverse sine of the input ``x``, element-wise.
+    Return the inverse sine of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -110,16 +117,16 @@ class asin(elementwise_unary):
 @register_op(doc_str="")
 class atan(elementwise_unary):
     """
-    Returns the inverse tangent of the input ``x``, element-wise.
+    Return the inverse tangent of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -137,16 +144,17 @@ class atan(elementwise_unary):
 @register_op(doc_str="")
 class atanh(elementwise_unary):
     """
-    Returns the inverse hyperbolic tangent values of the input ``x``, element-wise.
+    Return the inverse hyperbolic tangent values of the input
+    ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -164,16 +172,16 @@ class atanh(elementwise_unary):
 @register_op(doc_str="")
 class ceil(elementwise_unary):
     """
-    Returns the ceil values of the input ``x``, element-wise.
+    Return the ceil values of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -197,14 +205,14 @@ class clip(Operation):
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
     alpha: const f32 (Required)
     beta: const f32 (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -231,16 +239,16 @@ class clip(Operation):
 @register_op(doc_str="")
 class cos(elementwise_unary):
     """
-    Returns cosine of ``x`` element-wise. Input domain is ``(-inf, inf)`` and output
-    range is ``[-1,1]``.
+    Return cosine of ``x`` element-wise. Input domain is ``(-inf, inf)`` and
+    output range is ``[-1,1]``.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], T>
+    tensor<[\*d], T>
 
     Attributes
     ----------
@@ -258,15 +266,16 @@ class cos(elementwise_unary):
 @register_op(doc_str="")
 class cosh(elementwise_unary):
     """
-    Returns hyperbolic cosine of the input ``x``, element-wise.
+    Return hyperbolic cosine of the input ``x``, element-wise.
+
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], T>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], T>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -284,16 +293,16 @@ class cosh(elementwise_unary):
 @register_op(doc_str="")
 class erf(elementwise_unary):
     """
-    Returns the gauss error function of the input ``x``, element-wise.
+    Return the gauss error function of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -305,22 +314,23 @@ class erf(elementwise_unary):
 
     @precondition(allow=VALUE)
     def value_inference(self):
-        return scipy.special.erf(self.x.val)
+        erf_vector_function = np.vectorize(math.erf)
+        return erf_vector_function(self.x.val)
 
 
 @register_op(doc_str="")
 class exp(elementwise_unary):
     """
-    Returns the exponential values of the input ``x``, element-wise.
+    Return the exponential values of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -338,16 +348,16 @@ class exp(elementwise_unary):
 @register_op(doc_str="")
 class exp2(elementwise_unary):
     """
-    Returns the exponential values of the input ``x``, element-wise.
+    Return the exponential values of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -365,16 +375,17 @@ class exp2(elementwise_unary):
 @register_op(doc_str="")
 class floor(elementwise_unary):
     """
-    Returns the floor of the input ``x``, element-wise.
+    Return the floor of the input ``x``, element-wise, the same as rounding
+    towards negative infinity.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -392,20 +403,20 @@ class floor(elementwise_unary):
 @register_op(doc_str="")
 class inverse(Operation):
     """
-    Returns the reciprocal value of the input ``x``, element-wise.
+    Return the reciprocal value of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
-    epsilon: const f32 (Optional, default=1e-4)
-        * this is a small constant that is added to the input, before taking its inverse,
-          for stability.
-        * y = 1 / (x + epsilon)
+    x: tensor<[\*d], T> (Required)
+    epsilon: const fp32 (Optional, default=1e-4)
+        * This is a small constant that is added to the input, before taking its
+          inverse, for stability.
+        * ``y = 1 / (x + epsilon)``.
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -414,9 +425,14 @@ class inverse(Operation):
 
     input_spec = InputSpec(
         x=ScalarOrTensorInputType(),
-        epsilon=FloatInputType(const=True, default=1e-4),
+        epsilon=FloatInputType(const=True, optional=True),
     )
 
+    def default_inputs(self):
+        return DefaultInputs(
+            epsilon=1e-4,
+            )
+
     def __init__(self, **kwargs):
         super(inverse, self).__init__(**kwargs)
 
@@ -431,19 +447,19 @@ class inverse(Operation):
 @register_op(doc_str="")
 class log(Operation):
     """
-    Returns the natural logarithm value of the input ``x``, element-wise.
+    Return the natural logarithm value of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
-    epsilon: const f32 (Optional, default=1e-45)
-        * this is a small constant that is added to the input, before taking log.
-        * y = log(x + epsilon)
+    x: tensor<[\*d], T> (Required)
+    epsilon: const fp32 (Optional, default=1e-45)
+        * This is a small constant that is added to the input, before taking log.
+        * ``y = log(x + epsilon)``.
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -452,9 +468,13 @@ class log(Operation):
 
     input_spec = InputSpec(
         x=ScalarOrTensorInputType(),
-        epsilon=FloatInputType(const=True, default=1e-45),
+        epsilon=FloatInputType(const=True, optional=True),
     )
 
+    def default_inputs(self):
+        return DefaultInputs(
+            epsilon=1e-45)
+
     def __init__(self, **kwargs):
         super(log, self).__init__(**kwargs)
 
@@ -469,16 +489,18 @@ class log(Operation):
 @register_op(doc_str="")
 class logical_not(elementwise_unary):
     """
-    Returns the value of NOT the input ``x``, element-wise.
+    Return the value of NOT the input ``x``, element-wise. (``1`` for true, ``0``
+    for false in numeric domain.) A numeric value ``t`` is evaluated to true
+    ``iff t != 0``.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -496,16 +518,16 @@ class logical_not(elementwise_unary):
 @register_op(doc_str="")
 class round(elementwise_unary):
     """
-    Returns the round value of the input ``x``, element-wise.
+    Return the round value of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -523,20 +545,20 @@ class round(elementwise_unary):
 @register_op(doc_str="")
 class rsqrt(Operation):
     """
-    Returns the reciprocal value of the square root of the input ``x``, element-wise.
+    Return the reciprocal value of the square root of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
-    epsilon: const f32 (Optional, default=1e-12)
-        * this is a small constant that is added to the input, before applying the rsqrt function,
-          for stability.
-        * y = 1 / sqrt(x + epsilon)
+    x: tensor<[\*d], T> (Required)
+    epsilon: const fp32 (Optional, default=1e-12)
+        * This is a small constant that is added to the input, before applying the
+          ``rsqrt`` function, for stability.
+        * ``y = 1 / sqrt(x + epsilon)``.
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -545,9 +567,14 @@ class rsqrt(Operation):
 
     input_spec = InputSpec(
         x=ScalarOrTensorInputType(),
-        epsilon=FloatInputType(const=True, default=1e-12),
+        epsilon=FloatInputType(const=True, optional=True),
     )
 
+    def default_inputs(self):
+        return DefaultInputs(
+            epsilon=1e-12,
+            )
+
     def __init__(self, **kwargs):
         super(rsqrt, self).__init__(**kwargs)
 
@@ -562,16 +589,16 @@ class rsqrt(Operation):
 @register_op(doc_str="")
 class sign(elementwise_unary):
     """
-    Returns the sign value of the input ``x``, element-wise.
+    Return the sign value of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -589,16 +616,16 @@ class sign(elementwise_unary):
 @register_op(doc_str="")
 class sin(elementwise_unary):
     """
-    Returns the sine value of the input ``x``, element-wise.
+    Return the sine value of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -616,16 +643,16 @@ class sin(elementwise_unary):
 @register_op(doc_str="")
 class sinh(elementwise_unary):
     """
-    Returns the hyperbolic sine value of the input ``x``, element-wise.
+    Return the hyperbolic sine value of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -647,12 +674,12 @@ class sqrt(elementwise_unary):
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -670,16 +697,16 @@ class sqrt(elementwise_unary):
 @register_op(doc_str="")
 class square(elementwise_unary):
     """
-    Returns the square value of the input ``x``, element-wise.
+    Return the square value of the input ``x``, element-wise.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -697,16 +724,17 @@ class square(elementwise_unary):
 @register_op(doc_str="")
 class tan(elementwise_unary):
     """
-    Returns the tangent value of the input ``x``, element-wise.
+    Return the tangent value of the input ``x``, element-wise. Both input and output
+    ranges are ``(-inf, inf)``.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -724,16 +752,17 @@ class tan(elementwise_unary):
 @register_op(doc_str="")
 class tanh(elementwise_unary):
     """
-    Returns the hyperbolic tangent value of the input ``x``, element-wise.
+    Return the hyperbolic tangent value of the input ``x``, element-wise. Both input
+    and output ranges are ``(-inf, inf)`` while output range is ``[-1,  1]``.
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -756,13 +785,13 @@ class threshold(Operation):
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
-    alpha: const f32 (Required)
+    x: tensor<[\*d], T> (Required)
+    alpha: const fp32 (Required)
 
     Returns
     -------
-    tensor<[*d], f32>
-        * a tensor of the same shape as ``x``.
+    tensor<[\*d], f32>
+        * A tensor of the same shape as ``x``.
 
     Attributes
     ----------
@@ -791,18 +820,18 @@ class cast(Operation):
 
     Parameters
     ----------
-    x: tensor<[*d], T> (Required)
+    x: tensor<[\*d], T> (Required)
     dtype: const str (Required)
-        * Can be one of the following types: int32, int64, fp32, fp64.
+        * Can be one of the following types: ``int32``, ``int64``, ``fp32``, ``fp64``.
 
     Returns
     -------
-    tensor<[*d], dtype>
-        * a tensor of the same shape as ``x``, with type ``dtype``.
+    tensor<[\*d], dtype>
+        * A tensor of the same shape as ``x``, with type ``dtype``.
 
     Attributes
     ----------
-    T: i32, i64, fp32, fp64, bool.
+    T: i32, i64, fp16, fp32, fp64, bool.
     """
 
     input_spec = InputSpec(
@@ -816,6 +845,7 @@ class cast(Operation):
         type_map = {
             "int32": types.int32,
             "int64": types.int64,
+            "fp16": types.fp16,
             "fp32": types.fp32,
             "fp64": types.fp64,
             "bool": types.bool,
@@ -833,11 +863,12 @@ class cast(Operation):
         ret_shape = self.x.shape
         return types.tensor(type_map[self.dtype.val], ret_shape)
 
-    @precondition(allow=VALUE)
+    @precondition(allow=VALUE | SYMBOL)
     def value_inference(self):
         type_map = {
             "int32": np.int32,
             "int64": np.int64,
+            "fp16": np.float16,
             "fp32": np.float32,
             "fp64": np.float64,
             "bool": np.bool,
@@ -849,6 +880,12 @@ class cast(Operation):
                 "Provided {}".format(type_map.keys(), self.dtype.val)
             )
 
+        if self.x.val is None:
+            if self.x.sym_val is not None and not is_symbolic(self.x.sym_val) and len(self.x.sym_val.shape) == 1:
+                result = [np.array(val).astype(dtype=type_map[self.dtype.val]).item() if not is_symbolic(val) else val for val in self.x.sym_val]
+                return np.array(result)
+            return None
+
         if not types.is_tensor(self.x.sym_type):
             return self.x.val.astype(dtype=type_map[self.dtype.val])
         else:
