@@ -2,59 +2,90 @@
 #
 #  Use of this source code is governed by a BSD-3-clause license that can be
 #  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause
-
-from coremltools.converters.mil.mil import SYMBOL, VALUE
-from coremltools.converters.mil.mil.types.symbolic import is_compatible_symbolic_vector
-from ._op_reqs import *
+import numpy as np
 import numbers
 
+from coremltools.converters.mil.mil import Operation, types, SYMBOL, VALUE
+from coremltools.converters.mil.mil.input_type import (
+    DefaultInputs,
+    InputSpec,
+    IntInputType,
+    IntTensorInputType,
+    TensorInputType,
+    StringInputType,
+)
+from coremltools.converters.mil.mil.operation import precondition
+from coremltools.converters.mil.mil.ops.defs._op_reqs import register_op
+from coremltools.converters.mil.mil.types.symbolic import is_compatible_symbolic_vector, is_symbolic
+
 
 @register_op(doc_str="")
 class gather(Operation):
     """
-    Gather slices from input ``x`` along dimension ``axis`` according to ``indices``, similar to ``tf.gather``.
+    Gather slices from input ``x`` along dimension ``axis`` according to ``indices``,
+    similar to `tf.gather <https://www.tensorflow.org/api_docs/python/tf/gather>`_.
 
-    If ``indices`` is scalar (0-D):
-    ``output[p_0, ..., p_{axis-1},            ,p_{axis+1}, ..., p_{rank(x)-1}]
-    = x[p_0, ..., p_{axis-1}, indices, p_{axis+1}, ..., p_{rank(x)-1}]``
-    where ``rank(x)`` is the rank of ``x``.\n
-    The output has rank ``rank(x) - 1``.
+    * If ``indices`` is scalar (0-D):
+
+    .. math::
+       output[p_0, ..., p_{axis-1}, ~~~~~~~~~~~~~~~~~~~~~~~~ p_{axis+1}, ..., p_{rank(x)-1}] =
+    .. math::
+       x[p_0, ..., p_{axis-1}, ~~~~~~~~~ indices, ~~~~~~~~ p_{axis+1}, ..., p_{rank(x)-1}]
+
+    Where ``rank(x)`` is the rank of ``x``. The ``output`` has rank ``rank(x) - 1``.
+
+    * If ``indices`` is 1-D tensor:
+
+    .. math::
+       output[p_0, ..., p_{axis-1}, ~~~~~~~~~~~~~ i, ~~~~~~~~~~~~~ p_{axis+1}, ..., p_{rank(*D)-1}] =
+    .. math::
+       x[p_0, ..., p_{axis-1}, ~~~~~~~~ indices[i], ~~~~~~~~ p_{axis+1}, ..., p_{rank(*D)-1}]
 
-    If indices is 1-D tensor:
-    ``output[p_0, ..., p_{axis-1},            i, p_{axis+1}, ..., p_{rank(*D)-1}]
-    = x[p_0, ..., p_{axis-1}, indices[i], p_{axis+1}, ..., p_{rank(*D)-1}]``.
     The output has rank ``rank(x)``.
 
-    In general,
-    ``output[p_0, ..., p_{axis-1},           i_0, ..., i_{M-1} , p_{axis+1}, ..., p_{rank(x)-1}]
-    = x[p_0, ..., p_{axis-1}, indices[i_0, ..., i_{M-1}], p_{axis+1}, ..., p_{rank(x)-1}]``.
+    * In general:
 
-    where ``M = rank(x)``.
+    .. math::
+       output[p_0, ..., p_{axis-1}, ~~~~~~~~ i_0, ..., i_{M-1}, ~~~~~~~~ p_{axis+1}, ..., p_{rank(x)-1}] =
+    .. math::
+       x[p_0, ..., p_{axis-1}, ~~~~~~~ indices[i_0, ..., i_{M-1}], ~~~~~~~ p_{axis+1}, ..., p_{rank(x)-1}]
+
+    Where ``M = rank(x)``.
 
     Parameters
     ----------
-    * x: tensor<*D,T> (Required)
-    * indices: tensor<*N,i32> (Required)
+    x: tensor<\*D,T> (Required)
+    indices: tensor<\*N,i32> (Required)
         * Indices values may be negative. More precisely, ``-D[axis]<= v < D[axis]`` for ``v`` in ``indices``.
-    * axis: const i32 (Optional)
-        * Default to ``0``.
+    axis: const i32 (Optional. Default=``0``)
+        * Negative axis is supported.
 
     Returns
     -------
-    tensor<*K,T>
-        * where ``K = D[:axis] + N + D[axis+1:]``.
+    tensor<\*K,T>
+        * Where ``K = D[:axis] + N + D[axis+1:]``.
 
     Attributes
     ----------
     T: fp32
+
+    References
+    ----------
+    See `tf.gather <https://www.tensorflow.org/api_docs/python/tf/gather>`_.
+
     """
 
     input_spec = InputSpec(
         x=TensorInputType(),
         indices=IntInputType(),
-        axis=IntInputType(const=True, default=0),
+        axis=IntInputType(const=True, optional=True),
     )
 
+    def default_inputs(self):
+        return DefaultInputs(
+            axis=0,
+            )
+
     def __init__(self, **kwargs):
         super(gather, self).__init__(**kwargs)
 
@@ -101,53 +132,60 @@ class gather(Operation):
 @register_op(doc_str="")
 class scatter(Operation):
     """
-    Scatter ``updates`` to ``data`` at locations ``indices`` at dimension ``axis`` by operation ``mode``.
-    ``mode == update`` for instance.
-    ``
-    output[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D] =
-    updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
-    `` for ``i`` in ``[0, len(indices)]``.
-    Otherwise
-    ``
-    output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
-    data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
-    ``
+    Scatter ``updates`` to ``data`` at locations ``indices`` at dimension ``axis``
+    by operation ``mode``.
 
-    ``mode == add`` for instance.
-    ``
-    output[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D] =
-    updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D] +
-    x[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D]
-    `` for ``i`` in ``[0, len(indices)]
-    Otherwise
-    ``
-    output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
-    data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
-    ``
+    Example: ``mode == update``.
+
+    * For ``i`` in ``[0, len(indices)]``:
+
+    .. math::
+       output[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D] =
+    .. math::
+       updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
+
+    * For ``j! = i``:
+
+    .. math::
+       output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
+    .. math::
+       data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
+
+    Example: ``mode == add``.
+
+    * For ``i`` in ``[0, len(indices)]``:
+
+    .. math::
+       output[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D] =
+    .. math::
+       updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D] +
+    .. math::
+       x[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D]
+
+    * For ``j! = i``:
+
+    .. math::
+       output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
+    .. math::
+       data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
 
     Parameters
     ----------
-    data: tensor<*D, T> (Required)
+    data: tensor<\*D, T> (Required)
     indices: tensor<[C],T> (Required)
-        * 1D tensor.
-    updates: tensor<*K, T> (Required)
+        * 1-D tensor.
+    updates: tensor<\*K, T> (Required)
         * ``K = data.shape[:axis] + [len(indices)] + data.shape[axis+1:]``.
     axis: const i32 (Optional)
         * Default to ``0``.
     mode: const string (Optional)
-        * Can be the following modes:
-            * update
-            * add
-            * sub
-            * mul
-            * div
-            * max
-            * min
+        * Can be the following modes: ``update``, ``add``, ``sub``, ``mul``,
+          ``div``, ``max``, ``min``.
 
     Returns
     -------
-    tensor<*D, T>
-        * with the same type and shape as input ``x``.
+    tensor<\*D, T>
+        * With the same type and shape as input ``x``.
 
     Attributes
     ----------
@@ -158,10 +196,16 @@ class scatter(Operation):
         data=TensorInputType(),
         indices=IntTensorInputType(),
         updates=TensorInputType(),
-        axis=IntInputType(const=True, default=0),
-        mode=StringInputType(const=True, default="add"),
+        axis=IntInputType(const=True, optional=True),
+        mode=StringInputType(const=True, optional=True),
     )
 
+    def default_inputs(self):
+        return DefaultInputs(
+            axis=0,
+            mode="add",
+            )
+
     def __init__(self, **kwargs):
         super(scatter, self).__init__(**kwargs)
 
@@ -178,7 +222,15 @@ class scatter(Operation):
         expected_updates_shape = (
             self.data.shape[:axis] + self.indices.shape + self.data.shape[axis + 1 :]
         )
-        np.testing.assert_equal(self.updates.shape, np.array(expected_updates_shape))
+
+        err = "Updates shape {} is incorrect. It should be {}.".format(self.updates.shape, expected_updates_shape)
+        if len(self.updates.shape) == len(expected_updates_shape):
+            for dim1, dim2 in zip(self.updates.shape, expected_updates_shape):
+                if not is_symbolic(dim1) and not is_symbolic(dim2):
+                    if dim1 != dim2:
+                        raise ValueError(err)
+        else:
+            raise ValueError(err)
 
         return self.data.sym_type
 
@@ -187,24 +239,24 @@ class scatter(Operation):
 class gather_along_axis(Operation):
     """
     Take the values along ``axis`` at locations ``indices``.
-    ``
-    idx = indices[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D],
-    output[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D] =
-    = x[p_0, ..., p_{axis-1}, idx, p_{axis+1}, ..., p_D]
-    ``
+
+    .. math::
+       idx = indices[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
+    .. math::
+       output[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D] = = x[p_0, ..., p_{axis-1}, idx, p_{axis+1}, ..., p_D]
 
     Parameters
     ----------
-    x: tensor<*D, T> (Required)
-    indices: tensor<*K, T> (Required)
+    x: tensor<\*D, T> (Required)
+    indices: tensor<\*K, T> (Required)
         * ``rank(indices) == rank(x)``.
     axis: const i32 (Optional):
         * Default to ``0``.
 
     Returns
     -------
-    tensor<*D, T>:
-        * output tensor has the same shape as ``indices``.
+    tensor<\*D, T>:
+        * Output tensor has the same shape as ``indices``.
 
     Attributes
     ----------
@@ -214,9 +266,14 @@ class gather_along_axis(Operation):
     input_spec = InputSpec(
         x=TensorInputType(),
         indices=IntTensorInputType(),
-        axis=IntInputType(const=True, default=0),
+        axis=IntInputType(const=True, optional=True),
     )
 
+    def default_inputs(self):
+        return DefaultInputs(
+            axis=0,
+            )
+
     def __init__(self, **kwargs):
         super(gather_along_axis, self).__init__(**kwargs)
 
@@ -257,56 +314,65 @@ class gather_along_axis(Operation):
 @register_op(doc_str="")
 class scatter_along_axis(Operation):
     """
-    Scatter ``updates`` to ``data`` at locations ``indices`` at dimension ``axis`` by operation ``mode``.
-    ``mode == update`` for instance.
-    ``
-    idx = indices[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
-    output[p_0, ..., p_{axis-1}, idx, p_{axis+1}, ..., p_D] =
-    updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
-    `` for ``i`` in ``[0, len(indices)]``.
-    Otherwise
-    ``
-    output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
-    data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
-    ``
+    Scatter ``updates`` to ``data`` at locations ``indices`` at dimension ``axis``
+    by operation ``mode``.
 
-    ``mode == add`` for instance.
-    ``
-    idx = indices[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
-    output[p_0, ..., p_{axis-1}, idx, p_{axis+1}, ..., p_D] =
-    updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D] +
-    x[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D]
-    `` for ``i`` in ``[0, len(indices)]
-    Otherwise
-    ``
-    output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
-    data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
-    ``
+    Example: ``mode == update``.
+
+    * For ``i`` in ``[0, len(indices)]``:
+
+    .. math::
+       idx = indices[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
+    .. math::
+       output[p_0, ..., p_{axis-1}, idx, p_{axis+1}, ..., p_D] =
+    .. math::
+       updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
+
+    * For ``j! = i``:
+
+    .. math::
+       output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
+    .. math::
+       data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
+
+    Example: ``mode == add``.
+
+    * For ``i`` in ``[0, len(indices)]``:
+
+    .. math::
+       idx = indices[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D]
+    .. math::
+       output[p_0, ..., p_{axis-1}, idx, p_{axis+1}, ..., p_D] =
+    .. math::
+       updates[p_0, ..., p_{axis-1}, i, p_{axis+1}, ..., p_D] +
+    .. math::
+       x[p_0, ..., p_{axis-1}, indice[i], p_{axis+1}, ..., p_D]
+
+    * For ``j! = i``:
+
+    .. math::
+       output[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D] =
+    .. math::
+       data[p_0, ..., p_{axis-1}, j, p_{axis+1}, ..., p_D]
 
     Parameters
     ----------
-    data: tensor<*D, T> (Required)
-    indices: tensor<*K,T> (Required)
+    data: tensor<\*D, T> (Required)
+    indices: tensor<\*K,T> (Required)
         * ``rank(indices) == rank(data)``.
-    updates: tensor<*K, T> (Required)
-        * must be the same shape as ``indices``.
+    updates: tensor<\*K, T> (Required)
+        * Must be the same shape as ``indices``.
     axis: const i32 (Optional)
         * Default to ``0``.
     mode: const string (Optional)
         * Default to ``add``.
-        * Can be the following modes:
-            * ``update``
-            * ``add``
-            * ``sub``
-            * ``mul``
-            * ``div``
-            * ``max``
-            * ``min``
+        * Can be the following modes: ``update``, ``add``, ``sub``, ``mul``,
+          ``div``, ``max``, ``min``.
 
     Returns
     -------
-    tensor<*D, T>
-        * with the same type and shape as input ``x``.
+    tensor<\*D, T>
+        * With the same type and shape as input ``x``.
 
     Attributes
     ----------
@@ -317,10 +383,16 @@ class scatter_along_axis(Operation):
         data=TensorInputType(),
         indices=IntTensorInputType(),
         updates=TensorInputType(),
-        axis=IntInputType(const=True, default=0),
-        mode=StringInputType(const=True, default="add"),
+        axis=IntInputType(const=True, optional=True),
+        mode=StringInputType(const=True, optional=True),
     )
 
+    def default_inputs(self):
+        return DefaultInputs(
+            axis=0,
+            mode="add",
+            )
+
     def __init__(self, **kwargs):
         super(scatter_along_axis, self).__init__(**kwargs)
 
@@ -357,30 +429,40 @@ class scatter_along_axis(Operation):
 @register_op(doc_str="")
 class gather_nd(Operation):
     """
-    Gather slices from ``x`` according to ``indices``, similar to ``tf.gather_nd``.
+    Gather slices from ``x`` according to ``indices``, similar to `tf.gather_nd <https://www.tensorflow.org/api_docs/python/tf/gather_nd>`_.
 
-    ``indices`` is an K-dim tensor, where ``indices[i_0,...,i_{K-2}]`` defines a slice of ``x``:
+    The ``indices`` is a K-dim tensor, where ``indices[i_0,...,i_{K-2}]`` defines a slice
+    of ``x``:
 
-    ``output[i_0, ..., i_{K-2}]= x[indices[i_0, ..., i_{K-2}]]``
+    .. math::
+       output[i_0, ..., i_{K-2}]= x[indices[i_0, ..., i_{K-2}]]
 
-    where ``K = rank(indices)`` , and ``x[indices[i_0, ..., i_{K-2}]]`` has rank ``rank(x) - indices.shape[-1]``.
+    Where ``K = rank(indices)`` and ``x[indices[i_0, ..., i_{K-2}]]`` has rank
+    ``rank(x) - indices.shape[-1]``.
 
     Parameters
     ----------
-    * x: tensor<*D,T>  (Required)
-    * indices: tensor<*K,i32> (Required)
+    x: tensor<\*D,T> (Required)
+    indices: tensor<\*K,i32> (Required)
 
     Returns
     -------
-    * tensor<*V,T>
-        * V = K[:-1] + D[K[-1]:], where ``D = x.shape`` and ``K = indices.shape``.
+    tensor<\*V,T>
+        * ``V = K[:-1] + D[K[-1]:]``, where ``D = x.shape`` and ``K = indices.shape``.
 
     Attributes
     ----------
     T: fp32
+
+    References
+    ----------
+    See `tf.gather_nd <https://www.tensorflow.org/api_docs/python/tf/gather_nd>`_.
     """
 
-    input_spec = InputSpec(x=TensorInputType(), indices=IntTensorInputType(),)
+    input_spec = InputSpec(
+        x=TensorInputType(),
+        indices=IntTensorInputType(),
+        )
 
     def __init__(self, **kwargs):
         super(gather_nd, self).__init__(**kwargs)
@@ -395,37 +477,38 @@ class gather_nd(Operation):
 @register_op(doc_str="")
 class scatter_nd(Operation):
     """
-    Scatter ``updates`` to ``data`` at locations ``indices``.\n
+    Scatter ``updates`` to ``data`` at locations ``indices``.
 
-    ``indices`` is an K-dim tensor, where ``indices[i_0,...,i_{K-2}]`` defines a slice of ``data``,
-    where ``K = rank(indices)`` , and ``data[indices[i_0, ..., i_{K-2}]]`` has rank ``rank(data) - indices.shape[-1]``.\n
+    The ``indices`` is a K-dim tensor, where ``indices[i_0,...,i_{K-2}]`` defines a
+    slice of ``data``, ``K = rank(indices)``, and ``data[indices[i_0, ..., i_{K-2}]]``
+    has rank ``rank(data) - indices.shape[-1]``.
 
-    Take ``mode == update`` for instance:
-    ``output`` is set to ``data`` initially, and the op updates ``output`` as following:
-    ``output[indices[i_0, ..., i_{K-2}]]= updates[indices[i_0, ..., i_{K-2}]]``,
-    One more example for ``mode == add``, the update rule with be:
-    ``output[indices[i_0, ..., i_{K-2}]] += updates[indices[i_0, ..., i_{K-2}]]``
+    * Example: ``mode == update``: The ``output`` is set to ``data`` initially, and
+      the op updates ``output`` as follows:
+
+    .. math::
+       output[indices[i_0, ..., i_{K-2}]]= updates[indices[i_0, ..., i_{K-2}]]
+
+    * Example: ``mode == add``. The update rule is:
+
+    .. math::
+       output[indices[i_0, ..., i_{K-2}]] += updates[indices[i_0, ..., i_{K-2}]]
 
     Parameters
     ----------
-    data: tensor<*D,T>  (Required)
-    indices: tensor<*K,i32> (Required)
-    updates: tensor<*K, T> (Required)
-        * must be the shape as ``K[:-1]+data.shape[K[-1]:]``.
+    data: tensor<\*D,T> (Required)
+    indices: tensor<\*K,i32> (Required)
+    updates: tensor<\*K, T> (Required)
+        * Must be the shape as ``K[:-1]+data.shape[K[-1]:]``.
     mode: const string (Optional)
         * Default to ``add``.
-        * Can be the following modes:
-            * ``update``
-            * ``add``
-            * ``sub``
-            * ``mul``
-            * ``div``
-            * ``max``
-            * ``min``
+        * Can be the following modes: ``update``, ``add``, ``sub``, ``mul``,
+          ``div``, ``max``, ``min``.
+
     Returns
     -------
-    tensor<*D,T>
-        * a tensor with the same shape and type as ``data``.
+    tensor<\*D,T>
+        * A tensor with the same shape and type as ``data``.
 
     Attributes
     ----------
@@ -436,9 +519,14 @@ class scatter_nd(Operation):
         data=TensorInputType(),
         indices=IntTensorInputType(),
         updates=TensorInputType(),
-        mode=StringInputType(const=True, default="add"),
+        mode=StringInputType(const=True, optional=True),
     )
 
+    def default_inputs(self):
+        return DefaultInputs(
+            mode="add",
+            )
+
     def __init__(self, **kwargs):
         super(scatter_nd, self).__init__(**kwargs)
 
