@@ -171,7 +171,7 @@ class TestScatterAlongAxis:
         v = mb.scatter_along_axis(
             data=x, indices=indices, updates=updates, axis=0, mode="update"
         )
-        assert is_close(np.array([[1, 6, 10], [8, 9, 7]], dtype=np.float32), v.val)
+        np.testing.assert_allclose(np.array([[1, 6, 10], [8, 9, 7]], dtype=np.float32), v.val, atol=1e-04, rtol=1e-05)
 
     @pytest.mark.parametrize(
         "use_cpu_only, backend, rank_axis",
@@ -225,7 +225,6 @@ class TestScatterAlongAxis:
 
 
 class TestScatterNd:
-    # TODO: <rdar://problem/59737282> [MIL] Scatter and ScatterNd in tensoflow
     @pytest.mark.parametrize(
         "use_cpu_only, backend", itertools.product([True, False], backends,)
     )
@@ -359,6 +358,9 @@ class TestGather:
                 mb.gather(x=x, indices=indices, axis=-2),
                 mb.gather(x=x, indices=indices, axis=-1),
                 mb.gather(x=x, indices=indices),
+                # mb.gather(x=x, indices=1), #shape of scalar indices is incorrect.
+                # mb.gather(x=x, indices=1, axis=1), #Scalar index passes on axis=0 but fails on axis=1,
+                # Need to handle rank 0 correctly, rdar://73160449
             ]
 
         expected_output_types = [
@@ -367,6 +369,7 @@ class TestGather:
             (2, 3, types.fp32),
             (2, 2, types.fp32),
             (2, 3, types.fp32),
+            # (3, types.fp32),
         ]
 
         expected_outputs = [
@@ -375,6 +378,46 @@ class TestGather:
             np.array([[4, 5, 6], [1, 2, 3]], dtype=np.float32),
             np.array([[2, 1], [5, 4]], dtype=np.float32),
             np.array([[4, 5, 6], [1, 2, 3]], dtype=np.float32),
+            # np.array([4, 5, 6], dtype=np.float32),
+        ]
+
+        run_compare_builder(
+            build,
+            input_placeholders,
+            input_values,
+            expected_output_types,
+            expected_outputs,
+            use_cpu_only=use_cpu_only,
+            frontend_only=False,
+            backend=backend,
+        )
+
+    @pytest.mark.parametrize(
+        "use_cpu_only, backend", itertools.product([True, False], backends,)
+    )
+    def test_embedding_builder_to_backend_smoke(self, use_cpu_only, backend):
+        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
+        indices = np.array([1, 0], dtype=np.int32)
+        input_placeholders = {
+            "indices": mb.placeholder(shape=indices.shape, dtype=types.int32),
+        }
+
+        input_values = {"indices": indices}
+
+        def build(indices):
+            return [
+                mb.gather(x=x, indices=indices, axis=0),
+                mb.gather(x=x, indices=indices, axis=-2),
+            ]
+
+        expected_output_types = [
+            (2, 3, types.fp32),
+            (2, 3, types.fp32),
+        ]
+
+        expected_outputs = [
+            np.array([[4, 5, 6], [1, 2, 3]], dtype=np.float32),
+            np.array([[4, 5, 6], [1, 2, 3]], dtype=np.float32),
         ]
 
         run_compare_builder(
@@ -432,7 +475,7 @@ class TestGather:
         x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
         indices = np.array([1, 0], dtype=np.int32)
         v = mb.gather(x=x, indices=indices, axis=-1)
-        assert is_close(np.array([[2, 1], [5, 4]], dtype=np.float32), v.val)
+        np.testing.assert_allclose(np.array([[2, 1], [5, 4]], dtype=np.float32), v.val, atol=1e-04, rtol=1e-05)
 
 
 class TestGatherAlongAxis:
@@ -440,6 +483,10 @@ class TestGatherAlongAxis:
         "use_cpu_only, backend", itertools.product([True, False], backends,)
     )
     def test_builder_to_backend_smoke(self, use_cpu_only, backend):
+
+        if backend[0] == "mlprogram" and not use_cpu_only:
+            pytest.xfail("rdar://80710279 (Crash in GatherAlongAxis unit test on MIL GPU context)")
+
         x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
         indices = np.array([[1, 0, 1], [1, 1, 0]], dtype=np.int32)
         input_placeholders = {
@@ -490,17 +537,20 @@ class TestGatherAlongAxis:
         x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
         indices = np.array([[1, 0, 1], [0, 0, 1]], dtype=np.int32)
         v = mb.gather_along_axis(x=x, indices=indices, axis=0)
-        assert is_close(np.array([[4, 2, 6], [1, 2, 6]], dtype=np.float32), v.val)
+        np.testing.assert_allclose(np.array([[4, 2, 6], [1, 2, 6]], dtype=np.float32), v.val, atol=1e-04, rtol=1e-05)
 
     @pytest.mark.parametrize(
-        "use_cpu_only, backend, rank_axis",
+        "use_cpu_for_conversion, backend, rank_axis",
         itertools.product(
             [True, False],
             backends,
             [(rank, axis) for rank in range(1, 5) for axis in range(-rank, rank)],
         ),
     )
-    def test_builder_to_backend_programmatic(self, use_cpu_only, backend, rank_axis):
+    def test_builder_to_backend_programmatic(self, use_cpu_for_conversion, backend, rank_axis):
+        if backend[0] == "mlprogram" and not use_cpu_for_conversion:
+            pytest.xfail("rdar://78343225 ((MIL GPU) Core ML Tools Unit Test failures [numerical error])")
+
         rank, axis = rank_axis
         x_shape = np.random.randint(low=2, high=8, size=rank)
         indices_shape = np.copy(x_shape)
@@ -530,9 +580,10 @@ class TestGatherAlongAxis:
             input_values,
             expected_output_types,
             expected_output,
-            use_cpu_only=use_cpu_only,
+            use_cpu_only=use_cpu_for_conversion,
             frontend_only=False,
             backend=backend,
+            use_cpu_for_conversion=use_cpu_for_conversion,
         )
 
 
