@@ -2,16 +2,30 @@
 #
 #  Use of this source code is governed by a BSD-3-clause license that can be
 #  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause
+from pathlib import Path
 
+import os
 import logging
 import numpy as np
 import copy
+import re
 
 from coremltools.converters.mil.mil import Program, Function
+from coremltools.converters.mil.mil.passes.quantization_passes import AbstractQuantizationPass
 from coremltools.converters.mil.mil.passes.pass_registry import PASS_REGISTRY
 from coremltools._deps import _IS_MACOS
+
+import coremltools as ct
 import PIL.Image
 
+np.random.seed(10)
+
+def _serialize_current_pytest(mlmodel):
+    class_name = os.environ.get('PYTEST_CURRENT_TEST').split("::")[1].strip()
+    test_name = "::".join(os.environ.get('PYTEST_CURRENT_TEST').split("::")[2:]).split("(call)")[0].strip()
+    mlpackage_path = "/tmp/pytest_failures/{}/{}/model.mlpackage".format(class_name,test_name)
+    Path(mlpackage_path).mkdir(parents=True, exist_ok=True)
+    mlmodel.save(mlpackage_path)
 
 def assert_op_count_match(program, expect, op=None, verbose=False):
     """
@@ -32,7 +46,7 @@ def assert_op_count_match(program, expect, op=None, verbose=False):
 
 
 def assert_model_is_valid(
-    program, inputs, backend="nn_proto", verbose=True, expected_output_shapes=None
+    program, inputs, backend=("neuralnetwork", "fp32"), verbose=True, expected_output_shapes=None
 ):
     """
     Assert Core ML model is valid.
@@ -42,13 +56,14 @@ def assert_model_is_valid(
     - input: str -> shape tuple. All program input names need to appear in str.
       shape tuple can only contain positive integers.
     """
+    # Avoid circular import
+    from coremltools.converters.mil.testing_reqs import ct
+
     input_dict = dict()
     for name, shape in inputs.items():
         input_dict[name] = np.random.rand(*shape)
 
-    # Avoid circular import
-    from coremltools.converters.mil.testing_reqs import ct
-    mlmodel = ct.convert(program, source="mil", convert_to=backend)
+    mlmodel = ct_convert(program, source="milinternal", convert_to=backend, useCPUOnly=True)
     assert mlmodel is not None
 
     if verbose:
@@ -75,6 +90,18 @@ def assert_same_output_shapes(prog1, prog2, func_name="main"):
     prog2_output_shapes = [o.shape for o in prog2[func_name].outputs]
     assert prog1_output_shapes == prog2_output_shapes
 
+def get_op_names_in_program(prog, func_name="main", skip_const_ops=True):
+    """
+    Return the operations names in prog[func_name],
+    in the same order as they are stored (topological)
+    """
+    op_names_in_program = []
+    for op in prog[func_name].operations:
+        if skip_const_ops:
+            if op.op_type == "const":
+                continue
+        op_names_in_program.append(op.name)
+    return op_names_in_program
 
 def get_op_types_in_program(prog, func_name="main", skip_const_ops=True):
     """
@@ -138,28 +165,7 @@ def to_tuple(v):
     return tuple(v)
 
 
-def is_close(expected, actual, atol=1e-04, rtol=1e-05):
-    """
-    expected, actual: np.array or python primitive (scalar)
-    rtol: relative tolerance. See numpy.isclose.
-    """
-
-    close = np.isclose(expected, actual, atol=atol, rtol=rtol)
-    if not np.all(close):
-        diff = expected - actual
-        num_not_close = np.sum(~close)
-        msg = "Values differ by L1 norm: {}. Num entries not close: {}/{}"
-        logging.error(msg.format(np.sum(np.abs(diff)), num_not_close, expected.size))
-        if num_not_close < 30:
-            logging.error("Differing entries:")
-            logging.error("Expected: {}".format(expected[~close]))
-            logging.error("Actual: {}".format(actual[~close]))
-            logging.error("Delta: {}".format(diff[~close]))
-        return False
-    return True
-
-
-def run_core_ml_predict(mlmodel, input_key_values, use_cpu_only=False):
+def run_core_ml_predict(mlmodel, input_key_values, use_cpu_only=True):
     for k, v in input_key_values.items():
         if isinstance(v, PIL.Image.Image):
             continue
@@ -169,12 +175,20 @@ def run_core_ml_predict(mlmodel, input_key_values, use_cpu_only=False):
             input_key_values[k] = np.array([v], dtype=np.float32)
     return mlmodel.predict(input_key_values, useCPUOnly=use_cpu_only)
 
+def _get_coreml_out_from_dict(out_dict, out_name):
+    if out_name in out_dict:
+        return out_dict[out_name]
+    elif re.sub("[^a-zA-Z0-9_]", "_", out_name) in out_dict:
+        return out_dict[re.sub("[^a-zA-Z0-9_]", "_", out_name)]
+    else:
+        raise KeyError("{} output not found in Core ML outputs".format(out_name))
 
 def compare_backend(
     mlmodel,
     input_key_values,
     expected_outputs,
-    use_cpu_only=False,
+    use_cpu_only=True,
+    dtype = "fp32",
     atol=1e-04,
     rtol=1e-05,
     also_compare_shapes=True,
@@ -192,6 +206,10 @@ def compare_backend(
         - use_cpu_only: True/False.
     """
     if _IS_MACOS:
+
+        if dtype not in ["fp32", "fp16"]:
+            raise ValueError("Unsupported dtype config")
+
         pred = run_core_ml_predict(mlmodel, input_key_values,
             use_cpu_only=use_cpu_only)
         if also_compare_shapes:
@@ -202,21 +220,16 @@ def compare_backend(
                 use_cpu_only=use_cpu_only,
                 pred=pred,
             )
-        if not use_cpu_only:
+        if not use_cpu_only or (dtype == "fp16"):
             atol = max(atol * 100.0, 5e-1)
             rtol = max(rtol * 100.0, 5e-2)
         for o, expected in expected_outputs.items():
-            msg = (
-                "Output {} differs. useCPUOnly={}.\nInput={}, "
-                + "\nExpected={}, \nOutput={}\n"
-            )
-            assert is_close(expected, pred[o], atol, rtol), msg.format(
-                o, use_cpu_only, input_key_values, expected, pred[o]
-            )
+            coreml_out = _get_coreml_out_from_dict(pred, o)
+            np.testing.assert_allclose(coreml_out, expected, atol=atol, rtol=rtol)
 
 
 def compare_shapes(
-    mlmodel, input_key_values, expected_outputs, use_cpu_only=False, pred=None
+    mlmodel, input_key_values, expected_outputs, use_cpu_only=True, pred=None
 ):
     """
     Inputs:
@@ -237,22 +250,72 @@ def compare_shapes(
             pred = run_core_ml_predict(mlmodel, input_key_values,
                 use_cpu_only)
         for o, expected in expected_outputs.items():
+            coreml_out = _get_coreml_out_from_dict(pred, o)
             msg = "Output: {}. expected shape {} != actual shape {}".format(
-                o, expected.shape, pred[o].shape
+                o, expected.shape, coreml_out.shape
             )
             # Core ML does not support scalar as output
             # remove this special case when support is added
-            if expected.shape == () and pred[o].shape == (1,):
+            if expected.shape == () and coreml_out.shape == (1,):
                 continue
-            assert pred[o].shape == expected.shape, msg
+            assert coreml_out.shape == expected.shape, msg
 
+def ct_convert(
+    program,
+    source = "auto",
+    inputs = None,
+    outputs = None,
+    classifier_config = None,
+    minimum_deployment_target = None,
+    convert_to = None,
+    compute_precision = None,
+    skip_model_load = False,
+    **kwargs,
+):
+
+    """
+    Overloaded ct.convert function with the only difference being in the argument `convert_to`
+    which in this overloaded call accepts a tuple of (target, dtype).
+    Ex: ("neuralnetwork", "fp32"), ("mlprogram", "fp16")
+    """
+
+    target, dtype = convert_to
+
+    if dtype not in ["fp32", "fp16"]:
+        raise ValueError("Unsupported dtype config")
+
+    compute_precision = ct.precision.FLOAT16 if dtype == "fp16" else ct.precision.FLOAT32
+    if target == "neuralnetwork":
+        compute_precision = None
+
+    mlmodel = ct.convert(
+                program,
+                source = source,
+                inputs = inputs,
+                outputs = outputs,
+                classifier_config = classifier_config,
+                minimum_deployment_target = minimum_deployment_target,
+                convert_to = target,
+                compute_precision = compute_precision,
+                skip_model_load = skip_model_load,
+                **kwargs
+    )
+
+    if os.environ.get("DEBUG_SAVE_MLMODEL", "0") == "1":
+        from coremltools.converters.mil.testing_utils import _serialize_current_pytest
+        _serialize_current_pytest(mlmodel)
+
+    return mlmodel
 
 def get_core_ml_prediction(
-    build, input_placeholders, input_values, use_cpu_only=False, 
-    backend="nn_proto"):
+    build, input_placeholders, input_values, use_cpu_only=True,
+    backend=("neuralnetwork", "fp32")):
     """
     Return predictions of the given model.
     """
+    # Avoid circular import
+    from coremltools.converters.mil.testing_reqs import ct
+
     program = Program()
     with Function(input_placeholders) as ssa_func:
         output_vars = build(**ssa_func.inputs)
@@ -263,21 +326,20 @@ def get_core_ml_prediction(
         ssa_func.set_outputs(output_vars)
         program.add_function("main", ssa_func)
 
-    # Avoid circular import
-    from coremltools.converters.mil.testing_reqs import ct
-    mlmodel = ct.convert(program, source="mil",
-        convert_to=backend, useCPUOnly=use_cpu_only)
+    mlmodel = ct_convert(program, source="milinternal",
+                         convert_to=backend, useCPUOnly=use_cpu_only)
     return mlmodel.predict(input_values, useCPUOnly=use_cpu_only)
 
 
-def apply_pass_and_basic_check(prog, pass_name):
+def apply_pass_and_basic_check(prog, pass_name, skip_output_name_check=False):
     """
     Apply pass to the program
     """
     prev_prog = copy.deepcopy(prog)
-    PASS_REGISTRY[pass_name](prog)
+    pass_name.apply(prog) if isinstance(pass_name, AbstractQuantizationPass) else PASS_REGISTRY[pass_name](prog)
     block = prog.functions["main"]
     prev_block = prev_prog.functions["main"]
-    assert_same_output_names(prev_prog, prog)
+    if not skip_output_name_check:
+        assert_same_output_names(prev_prog, prog)
     assert_same_output_shapes(prev_prog, prog)
     return prev_prog, prev_block, block
