@@ -6,26 +6,46 @@
 """
 Utilities for the entire package.
 """
+from coremltools.converters.mil.mil.passes.name_sanitization_utils import NameSanitizer as _NameSanitizer
+from coremltools.proto import Model_pb2 as _Model_pb2
 import math as _math
 import numpy as _np
 import os as _os
-import six as _six
-import warnings as _warnings
+import pathlib as _pathlib
 import sys as _sys
-from coremltools.proto import Model_pb2 as _Model_pb2
-from coremltools.models._deprecation import deprecated as _deprecated
-from .._deps import _HAS_SKLEARN
+import tempfile as _tempfile
+import warnings as _warnings
+from .._deps import _HAS_SCIPY
+from ..libmodelpackage import ModelPackage
 
-if _HAS_SKLEARN:
+
+_MLMODEL_EXTENSION = ".mlmodel"
+_MLPACKAGE_EXTENSION = ".mlpackage"
+
+
+try:
+    from ..libmodelpackage import ModelPackage
+except ModuleNotFoundError:
+    pass
+
+if _HAS_SCIPY:
     import scipy.sparse as _sp
 
 
 def _to_unicode(x):
-    if isinstance(x, _six.binary_type):
+    if isinstance(x, bytes):
         return x.decode()
     else:
         return x
 
+def _remove_invalid_keys(input_dict, model):
+    # make sure that input_dict does not contain an input name, which
+    # is not present in the list of model inputs
+    input_dict_keys = list(input_dict.keys())
+    model_input_names = set([inp.name for inp in model._spec.description.input])
+    for k in input_dict_keys:
+        if k not in model_input_names:
+            del input_dict[k]
 
 def save_spec(spec, filename, auto_set_specification_version=False):
     """
@@ -53,11 +73,17 @@ def save_spec(spec, filename, auto_set_specification_version=False):
     load_spec
     """
     name, ext = _os.path.splitext(filename)
+
+    is_package = False
+
     if not ext:
-        filename = "{}.mlmodel".format(filename)
+        filename = "{}{}".format(filename, _MLMODEL_EXTENSION)
+    elif ext == _MLPACKAGE_EXTENSION:
+        is_package = True
+    elif ext == _MLMODEL_EXTENSION:
+        is_package = False
     else:
-        if ext != ".mlmodel":
-            raise Exception("Extension must be .mlmodel (not {})".format(ext))
+        raise Exception("Extension must be {} or {} (not {})".format(_MLMODEL_EXTENSION, _MLPACKAGE_EXTENSION, ext))
 
     spec = spec.SerializeToString()
     if auto_set_specification_version:
@@ -74,9 +100,20 @@ def save_spec(spec, filename, auto_set_specification_version=False):
                 RuntimeWarning,
             )
 
-    with open(filename, "wb") as f:
+    specfile = filename
+    if is_package:
+        tempfile = _tempfile.NamedTemporaryFile(suffix=_MLMODEL_EXTENSION)
+        specfile = tempfile.name
+
+    with open(specfile, "wb") as f:
         f.write(spec)
 
+    if is_package:
+        package = ModelPackage(filename)
+        model_name = _pathlib.Path(filename).with_suffix('.mlmodel').name
+        
+        # Root file is copied into the model package. Changes to in-memory JSON is commited to disk when package goes out of scope.
+        package.replaceRootModel(specfile, model_name, "com.apple.CoreML", "CoreML Model Specification");
 
 def load_spec(filename):
     """
@@ -107,7 +144,11 @@ def load_spec(filename):
 
     spec = Model_pb2.Model()
 
-    with open(filename, "rb") as f:
+    specfile = filename
+    if ModelPackage.isValid(filename):
+        specfile = ModelPackage(filename).getRootModel().path()
+
+    with open(specfile, "rb") as f:
         contents = f.read()
         spec.ParseFromString(contents)
         return spec
@@ -190,14 +231,6 @@ def _wp_to_fp16wp(wp):
     wp.float16Value = _fp32_to_fp16_byte_array(wp.floatValue)
     del wp.floatValue[:]
 
-
-@_deprecated(
-    suffix="instead use 'coremltools.models.neural_network.quantization_utils.quantize_weights'."
-)
-def convert_neural_network_spec_weights_to_fp16(fp_spec):
-    return _convert_neural_network_spec_weights_to_fp16(fp_spec)
-
-
 def _convert_neural_network_spec_weights_to_fp16(fp_spec):
     from .neural_network.quantization_utils import _quantize_spec_weights
     from .neural_network.quantization_utils import (
@@ -208,13 +241,6 @@ def _convert_neural_network_spec_weights_to_fp16(fp_spec):
     return qspec
 
 
-@_deprecated(
-    suffix="instead use 'coremltools.models.neural_network.quantization_utils.quantize_weights'."
-)
-def convert_neural_network_weights_to_fp16(full_precision_model):
-    return _convert_neural_network_weights_to_fp16(full_precision_model)
-
-
 def _convert_neural_network_weights_to_fp16(full_precision_model):
     """
     Utility function to convert a full precision (float) MLModel to a
@@ -253,17 +279,15 @@ def _get_model(spec):
 def evaluate_regressor(model, data, target="target", verbose=False):
     """
     Evaluate a CoreML regression model and compare against predictions
-    from the original framework (for testing correctness of conversion)
+    from the original framework (for testing correctness of conversion).
 
     Parameters
     ----------
-    filename: list of str or list of MLModel
-        File path from which to load the MLModel from (OR) a loaded version of
-        MLModel.
+    model: MLModel or str
+        A loaded MLModel or a path to a saved MLModel
 
-    data: list of str or list of Dataframe
-        Test data on which to evaluate the models (dataframe,
-        or path to a .csv file).
+    data: Dataframe
+        Test data on which to evaluate the models
 
     target: str
        Name of the column in the dataframe that must be interpreted
@@ -294,7 +318,9 @@ def evaluate_regressor(model, data, target="target", verbose=False):
     error_squared = 0
 
     for index, row in data.iterrows():
-        predicted = model.predict(dict(row))[_to_unicode(target)]
+        input_dict = dict(row)
+        _remove_invalid_keys(input_dict, model)
+        predicted = model.predict(input_dict)[_to_unicode(target)]
         other_framework = row["prediction"]
         delta = predicted - other_framework
 
@@ -357,7 +383,9 @@ def evaluate_classifier(model, data, target="target", verbose=False):
     num_errors = 0
 
     for index, row in data.iterrows():
-        predicted = model.predict(dict(row))[_to_unicode(target)]
+        input_dict = dict(row)
+        _remove_invalid_keys(input_dict, model)
+        predicted = model.predict(input_dict)[_to_unicode(target)]
         other_framework = row["prediction"]
         if predicted != other_framework:
             num_errors += 1
@@ -405,6 +433,7 @@ def evaluate_classifier_with_probabilities(
 
     for _, row in data.iterrows():
         input_dict = {k: v for k, v in dict(row).items() if k != probabilities}
+        _remove_invalid_keys(input_dict, model)
         predicted_values = model.predict(input_dict)[_to_unicode(probabilities)]
         other_values = row[probabilities]
 
@@ -558,15 +587,41 @@ def rename_feature(
                 rename_outputs or (index < len(spec.pipeline.models)),
             )
 
+    # Rename for mlProgram
+    if spec.HasField("mlProgram"):
+        new_name_sanitized = _NameSanitizer().sanitize_name(new_name)
+        if new_name != new_name_sanitized:
+            raise ValueError("Input/output names for ML Program must be of the format [a-zA-Z_][a-zA-Z0-9_]*. "
+                             "That is, it must start with a letter and only contain numerals, underscore or letters. "
+                             "Provided feature name, \"{}\" does not satisfy these requirements.".format(new_name))
+        mil = spec.mlProgram
+        for function in mil.functions.values():
+            for name_value_type in function.inputs:
+                if name_value_type.name == current_name:
+                    name_value_type.name = new_name
+            for block in function.block_specializations.values():
+                for i, out_name in enumerate(block.outputs):
+                    if out_name == current_name:
+                        block.outputs[i] = new_name
+                for op in block.operations:
+                    for argument in op.inputs.values():
+                        for binding in argument.arguments:
+                            if binding.HasField("name"):
+                                if binding.name == current_name:
+                                    binding.name = new_name
+                    for name_value_type in op.outputs:
+                        if name_value_type.name == current_name:
+                            name_value_type.name = new_name
+
 
 def _sanitize_value(x):
     """
     Performs cleaning steps on the data so various type comparisons can
     be performed correctly.
     """
-    if isinstance(x, _six.string_types + _six.integer_types + (float,)):
+    if isinstance(x, (str, int, float,)):
         return x
-    elif _HAS_SKLEARN and _sp.issparse(x):
+    elif _HAS_SCIPY and _sp.issparse(x):
         return x.todense()
     elif isinstance(x, _np.ndarray):
         return x
@@ -777,11 +832,12 @@ def _macos_version():
     version comparisons. On non-Macs, it returns an empty tuple.
     """
     if _is_macos():
-        import platform
-
-        ver_str = platform.mac_ver()[0]
-        return tuple([int(v) for v in ver_str.split(".")])
-
+        try:
+            import subprocess
+            ver_str = subprocess.run(["sw_vers", "-productVersion"], stdout=subprocess.PIPE).stdout.decode('utf-8').strip('\n')
+            return tuple([int(v) for v in ver_str.split(".")])
+        except:
+            raise Exception("Unable to detemine the macOS version")
     return ()
 
 
